import re
from typing import List, Set, Any
import pickle
import os

"""
    Directory help:
"""


def list_files_in_directory(path, extensions=None):
    """
    List files in a directory, optionally filtering by extensions.

    :param path: Directory to search
    :param extensions: None, a string like 'txt', or a list like ['txt', 'csv']
    :return: List of matching file paths
    :raises FileNotFoundError: If the directory does not exist
    """
    if not os.path.isdir(path):
        raise FileNotFoundError(f"Directory not found: {path}")

    if extensions:
        if isinstance(extensions, str):
            extensions = [extensions]
        extensions = tuple(ext.lower().lstrip(".") for ext in extensions)

    result = []
    for file in os.listdir(path):
        full_path = os.path.join(path, file)
        if os.path.isfile(full_path):
            if not extensions or file.lower().endswith(
                tuple(f".{ext}" for ext in extensions)
            ):
                result.append(file)

    return result


"""
    File saving:
"""


def save_obj_pickle(file_path: str, raw_data) -> None:
    """
    Saves raw_data to file_path via pickle
    """
    with open(file_path, "wb") as handle:
        pickle.dump(raw_data, handle)


def read_obj_pickle(file_path: str):
    """
    Loads data in file_path via pickle
    """
    if not os.path.isfile(file_path):
        raise FileNotFoundError(f"Pickle file not found: {file_path}")

    with open(file_path, "rb") as f:
        return pickle.load(f)


"""
    Python default types:
"""


def map_dicts_values_to_keys(dict1, dict2) -> dict:
    """
    creates a new dictionary, which maps from keys of dict1 to values of dict2,
        assuming that the values of dict1 are keys of dict2
    """
    return {k: dict2[v] for k, v in dict1.items() if v in dict2}


def is_decimal(string: str) -> bool:
    """
    Given a string, checks if it is a representation of a decimal and returns a boolean
    """
    return bool(re.match(r"^-?\d+(\.\d+)?$", string))


def intersect_sets(sets: List[Set[Any]]) -> Set[Any]:
    """
    Compute the intersection of a list of sets.

    Args:
        sets (List[Set[Any]]): A list of set objects.

    Returns:
        Set[Any]: The intersection of all sets in the list.
    """
    if not sets:
        return set()

    return set.intersection(*sets)



import pandas as pd

class GeneralCVSplit:
    """
    General class to split indices.
    """

    def __init__(self):
        raise NotImplementedError

    def split(self, indices):
        raise NotImplementedError


class RollingWindowSplit(GeneralCVSplit):
    """
        Implements a (fixed) rolling window splitter.
    """
    def __init__(self, train_size: int, test_size: int, gap: int = 0):
        self.train_size = train_size
        self.test_size = test_size
        self.gap = gap

        self.min_size = self.train_size + self.test_size + self.gap
        
    def split(self, indices: pd.Series):
        assert (
            len(indices) >= self.min_size
        ), "Error: Number of indices is less than the minimum index size for training, test, and gap."
        
        train_start_index = 0
        while train_start_index + self.min_size <= len(indices):
            end_train = train_start_index + self.train_size
            start_test = end_train + self.gap  
            end_test = start_test + self.test_size
            yield indices.iloc[train_start_index:end_train], indices.iloc[
                start_test:end_test
            ]
            train_start_index += 1



import numpy as np
from scipy.stats import norm

def normal_CI_2_sided_multiplier(alpha: float) -> float:
    """
    Given significance level alpha (e.g., 0.05), return the confidence interval multiplier.
    """
    return norm.ppf(1 - alpha / 2)

    
def compute_sharpe_w_cov(
    weights: np.ndarray, mu: np.ndarray, Sigma: np.ndarray
) -> list:
    """
    Given a 2d array of weights, where each column is a new set of weights, computes the
        Sharpe ratio with the mean and cov.
    """
    return [
        np.sum(weight * mu) / np.sqrt(weight @ Sigma @ weight.T) for weight in weights.T
    ]


def compute_port_utility(weight, gamma, mu, Sigma):
    return (weight.T @ mu) - gamma / 2 * (weight.T @ Sigma @ weight)


import pandas as pd
import numpy as np
from typing import Union

"""
    moneyline <-> probability:
"""


def moneyline_to_prob(s: Union[pd.Series, float]) -> Union[pd.Series, float]:
    return np.where(s > 0, 100 / (s + 100), -s / (100 - s))


def prob_to_moneyline(s: Union[pd.Series, float]) -> Union[pd.Series, float]:
    return np.where(s <= 0.5, 100 / s - 100, 100 * s / (s - 1))


def moneyline_to_prob_spread(s):
    return np.sum(moneyline_to_prob(s)) - 1


"""
    EV + Variance of Bonus bets, etc...
"""


def bonus_bet_ev(
    p: Union[pd.Series, float], spread: float = 0.02
) -> Union[pd.Series, float]:
    return p * (1 / (p + spread) - 1)


def bonus_bet_var(
    p: Union[pd.Series, float], spread: float = 0.02
) -> Union[pd.Series, float]:
    return (p - p**2) * (1 / (p + spread) - 1) ** 2


def profit_boost_ev(
    p: Union[pd.Series, float], boost: float = 0.3, spread: float = 0.06
) -> Union[pd.Series, float]:
    return p * ((1 / (p + spread) - 1) * (1 + boost) + 1) - 1


def profit_boost_var(
    p: Union[pd.Series, float], boost: float = 0.3, spread: float = 0.06
) -> Union[pd.Series, float]:
    return (p - p**2) * ((1 / (p + spread) - 1) * (1 + boost) + 1) ** 2


def no_sweat_ev(
    p: Union[pd.Series, float], bonus_p: float = 0.5, spread: float = 0.06
) -> Union[pd.Series, float]:
    return (
        p * (1 / (p + spread)) + (1 - p) * (bonus_bet_ev(p=bonus_p, spread=spread)) - 1
    )



import pandas as pd
import datetime as dt
from typing import Union

def YYYYMM_to_date(s: pd.Series) -> pd.Series:
    """
        converts series of integers of form YYYYMM into dates
    """
    return pd.to_datetime(s.astype(str), format='%Y%m').dt.date

def date_id_to_date(date_id: Union[int, pd.Series]):
    if isinstance(date_id, int):
        return dt.datetime.strptime(str(date_id), '%Y%m%d').date()
    else:
        return pd.to_datetime(date_id.astype(str), format='%Y%m%d').dt.date

def date_to_date_id(date: dt.date) -> int:
    return int(date.strftime('%Y%m%d'))

def get_friday_of_isocalendar(iso_year, iso_week):
    # Get the first day of the ISO week
    first_day_of_week = dt.date(iso_year, 1, 1) + dt.timedelta(weeks=iso_week-1)
    
    # Adjust to the correct day of the week (Friday is 4)
    # `first_day_of_week.weekday()` gives the weekday of the first day of the ISO week.
    # We need to adjust to make sure we get the Friday of that week.
    days_to_friday = (4 - first_day_of_week.weekday()) % 7
    friday = first_day_of_week + dt.timedelta(days=days_to_friday)
    return friday



import pandas as pd
import plotly.graph_objects as go
import plotly.colors as pc
import warnings
from scipy.stats import norm


def plot_mean_with_ci(
    means_df: pd.DataFrame,
    stds_df: pd.DataFrame,
    alpha: float = 1.96,
    title: str = "",
    xaxis_title: str = "",
    yaxis_title: str = "",
):
    """
    Plot each column of means_df with confidence intervals using Plotly.

    Parameters:
        means_df (pd.DataFrame): DataFrame of means, indexed by time or observation.
        stds_df (pd.DataFrame): DataFrame of standard deviations (same shape as means_df).
        alpha (float): Significance level for Normal two-sided CI
    """
    assert means_df.shape == stds_df.shape, "Shape of means and stds df do not match..."

    ci_multiplier = norm.ppf(1 - alpha / 2)

    fig = go.Figure()
    color_cycle = pc.qualitative.Plotly  # 10-color palette
    if len(means_df.columns) > 10:
        warnings.warn("More than 10 columns to plot - may duplicate colors.")

    for i, col in enumerate(means_df.columns):
        mean = means_df[col]
        std = stds_df[col]
        upper = mean + ci_multiplier * std
        lower = mean - ci_multiplier * std
        color = color_cycle[i % len(color_cycle)]

        # Add mean line
        fig.add_trace(
            go.Scatter(
                x=means_df.index,
                y=mean,
                mode="lines",
                name=f"{col}",
                line=dict(color=color),
                legendgroup=col,  # Group line + CI
                showlegend=True,  # Show only for the line
                visible=True,
            )
        )

        # Add confidence interval as filled area
        fig.add_trace(
            go.Scatter(
                x=pd.concat(
                    [means_df.index.to_series(), means_df.index.to_series()[::-1]]
                ),
                y=pd.concat([upper, lower[::-1]]),
                fill="toself",
                fillcolor=color,
                opacity=0.2,
                line=dict(color="rgba(255,255,255,0)"),
                hoverinfo="skip",
                legendgroup=col,  # Belongs to same group
                showlegend=False,
                name=f"{col} CI",
                visible=True,
            )
        )

    fig.update_layout(
        title=title,
        xaxis_title=xaxis_title,
        yaxis_title=yaxis_title,
    )

    fig.show()

    return fig



import pandas as pd
import numpy as np

"""
    Df utils:
"""

def strip_excess_spaces_df(df: pd.DataFrame) -> pd.DataFrame:
    """
        Returns dataframe with leading + trailing spaces stripped for string columns
        - from gpt, but note that this has high mem usage + slow effectiveness
    """
    df_cleaned = df.copy()
    for col in df.columns:
        if df[col].apply(lambda x: isinstance(x, str)).mean() > 0.1:
            df_cleaned[col] = df[col].str.rstrip()
    return df_cleaned


def str_to_float_df(df: pd.DataFrame) -> pd.DataFrame:
    """
        Try to convert all string columns into floats or ints, depending on pd.to_numeric 
    """
    df_converted = df.copy()
    for col in df_converted.columns:
        try:
            df_converted[col] = pd.to_numeric(df_converted[col], errors='raise')
        except Exception as e:
            pass  # Skip columns that can't be converted
    return df_converted

def format_size(size_bytes) -> str:
    """Convert bytes to a human-readable format."""
    for unit in ['B', 'KB', 'MB', 'GB', 'TB']:
        if size_bytes < 1024.0:
            return f"{size_bytes:,.2f} {unit}"
        size_bytes /= 1024.0
    return f"{size_bytes:,.2f} PB"
    
def print_np_arr_mem_usage(arr: np.array) -> None:
    """Print memory usage of numpy array in human-readable format."""
    if not isinstance(arr, np.ndarray):
        print(f"Input is not a numpy array.")
        return
    size_bytes = arr.nbytes
    print(f"Array size: {format_size(size_bytes)}")